// Extract a GPS coordinates dump and list it

// Map fetcher
// Map: http://www.mapblast.com/gif?&CT=41.711632:-73.932541:25000&IC=&W=1280&H=1024&FAM=mblast&LB=
// lat, lon, zoom

#include "config.h"

// Prevent make dep warnings
#if (defined(HAVE_IMAGEMAGICK) && defined(HAVE_GPS))

#include <stdio.h>
#include <pthread.h>
#ifdef HAVE_STDINT_H
#include <stdint.h>
#endif
#ifdef HAVE_INTTYPES_H
#include <inttypes.h>
#endif
#include <math.h>
#include <time.h>
#include "getopt.h"
#include <unistd.h>
#include <list>
#include <map>
#include <vector>
#include <algorithm>
#include <string>
#include <zlib.h>
#include <magick/api.h>

#include "gpsdump.h"

#define MAJOR 2
#define MINOR 2

// looky looky, image magick changed their api!
#if MagickLibVersion < 0x0545
#define OldMagickLib
#endif

const char url_template[] = "http://www.mapblast.com/gif?&CT=%f:%f:%ld&IC=&W=%d&H=%d&FAM=myblast&LB=%s";
const char download_template[] = "wget \"%s\" -O %s";

// Decay from absolute blue for multiple tracks
const uint8_t track_decay = 0x1F;
// Width of the track
const unsigned int track_width = 3;

// distance (in feet) before we throttle a network and discard it
const unsigned int horiz_throttle = 75000;

// Image scales we use
long int scales[] = { 1000, 2000, 5000, 10000, 20000, 30000, 50000, 60000, 70000, 75000, 80000,
85000, 90000, 95000, 100000, 125000, 150000, 200000, 300000, 500000, 750000, 1000000, 2000000,
3000000, 4000000, 5000000, 6000000, 7000000, 8000000, 9000000, 10000000, 15000000,
20000000, 25000000, 30000000, 35000000, 40000000, 0 };

// Image colors
const char *netcolors[] = {
    "#000000",
    "#FF0000", "#FF0072", "#FF00E5", "#D400FF",
    "#5D00FF", "#0087FF", "#00F2FF", "#00FF94",
    "#00FF2E", "#BBFF00", "#FFB200", "#FF6E00",
    "#FF6500", "#960000", "#96005F", "#640096",
    "#001E96", "#008E96", "#00963E", "#529600",
    "#968C00", "#963700", NULL
};

// strength colors
char **powercolors;
int power_steps;

// Origional
char *powercolors_Orig[] = {
    "#FF0000", "#FFD500", "#FFCC00",
    "#F2FF00", "#7BFF00", "#00FFB6",
    "#00FFFF", "#005DFF", "#A100FF",
    "#FA00FF"
};
const int power_steps_Orig = 10;

// Math progression
char *powercolors_Math[] = {
    "#FF0000", "#FF8000", "#FFFF00",
    "#80FF00", "#00FF00", "#00FF80",
    "#00FFFF", "#0080FF", "#0000FF",
    "#8000FF", "#FF00FF", "#FF0080"
};
const int power_steps_Math = 12;

// Weather Radar
char *powercolors_Radar[] = {
    "#50E350", "#39C339", "#208420",
    "#145A14", "#C8C832", "#DC961E",
    "#E61E1E", "#B31A17", "#811610",
};
const int power_steps_Radar = 9;

const int power_max = 255;
//const double power_constant = -2;
unsigned int map_width = 1280;
unsigned int map_height = 1024;

int draw_opacity, draw_track, draw_bounds, draw_range, draw_interpolated,
    draw_legend, convert_greyscale, keep_gif, draw_center, center_dot, verbose, interpolation_res;

int sample_points;

typedef struct gps_network {
    net_hdr header;

    double max_lat;
    double min_lat;
    double max_lon;
    double min_lon;
    double max_alt;
    double min_alt;

    int count;

    double avg_lat, avg_lon, avg_alt, avg_spd;

    double diagonal_distance, altitude_distance;

    // Index to the netcolors table
    int color_index;
};

typedef struct {
    int version;
    double lat;
    double lon;
    int power;

    unsigned int x, y;
} track_data;

// Array of network track arrays
unsigned int num_tracks;
vector< vector<track_data> > track_vec;
// bssid map
map<string, gps_network> net_map;
// Global average for map scaling
gps_network global_map_avg;
// Mapquest map scale
long scale;
// Center lat/lon for map
double map_avg_lat, map_avg_lon;
// Do we have any power data?
int power_data;

// Threads, locks, and graphs to hold the power
pthread_t *mapthread;
pthread_mutex_t power_lock;
pthread_mutex_t print_lock;
unsigned int numthreads = 1;
unsigned int *power_map;
unsigned int power_pos = 0;
pthread_mutex_t power_pos_lock;
int *power_input_map;

// Forward prototypes
string Mac2String(uint8_t *mac, char seperator);
double rad2deg(double x);
double earth_distance(double lat1, double lon1, double lat2, double lon2);
double calcR (double lat);
void calcxy (double *posx, double *posy, double lat, double lon, double pixelfact,
             double zero_lat, double zero_long);
long int BestMapScale(double tlat, double tlon, double blat, double blon);
int ProcessGPSFile(char *in_fname);
void AssignNetColors();
void ProcessNetData(int in_printstats);
void DrawNetTracks(Image *in_img, DrawInfo *in_di);
void DrawNetCircles(Image *in_img, DrawInfo *in_di);
void DrawNetBoundRects(Image *in_img, DrawInfo *in_di, int in_fill);
void DrawLegend(Image *in_img, DrawInfo *in_di);
void DrawNetCenterDot(Image *in_img, DrawInfo *in_di, int in_dotsize);
int InverseWeight(int in_x, int in_y, int in_fuzz, double in_scale);
void DrawNetPower(Image *in_img, DrawInfo *in_di);

string Mac2String(uint8_t *mac, char seperator) {
    char tempstr[MAC_STR_LEN];

    // There must be a better way to do this...
    if (seperator != '\0')
        snprintf(tempstr, MAC_STR_LEN, "%02X%c%02X%c%02X%c%02X%c%02X%c%02X",
                 mac[0], seperator, mac[1], seperator, mac[2], seperator,
                 mac[3], seperator, mac[4], seperator, mac[5]);
    else
        snprintf(tempstr, MAC_STR_LEN, "%02X%02X%02X%02X%02X%02X",
                 mac[0], mac[1], mac[2],
                 mac[3], mac[4], mac[5]);

    string temp = tempstr;
    return temp;
}


// Faust Code to convert rad to deg and find the distance between two points
// on the globe.  Thanks, Faust.
//const float M_PI = 3.14159;
double rad2deg(double x) {
     return x*M_PI/180.0;
}

double earth_distance(double lat1, double lon1, double lat2, double lon2) {
    /*
    const double radius = 6378150.0; //meters

    //convert to rads
    double x1 = radius * cos(rad2deg(lon1)) * sin(rad2deg(90-lat1));
    double x2 = radius * cos(rad2deg(lon2)) * sin(rad2deg(90-lat2));
    double y1 = radius * sin(rad2deg(lon1)) * sin(rad2deg(90-lat1));
    double y2 = radius * sin(rad2deg(lon2)) * sin(rad2deg(90-lat2));
    double z1 = radius * cos(rad2deg(90-lat1));
    double z2 = radius * cos(rad2deg(90-lat2));
    double a = acos((x1*x2 + y1*y2 + z1*z2)/pow(radius,2));
    return radius * a;
    */

    double x1 = calcR(lat1) * cos(rad2deg(lon1)) * sin(rad2deg(90-lat1));
    double x2 = calcR(lat2) * cos(rad2deg(lon2)) * sin(rad2deg(90-lat2));
    double y1 = calcR(lat1) * sin(rad2deg(lon1)) * sin(rad2deg(90-lat1));
    double y2 = calcR(lat2) * sin(rad2deg(lon2)) * sin(rad2deg(90-lat2));
    double z1 = calcR(lat1) * cos(rad2deg(90-lat1));
    double z2 = calcR(lat2) * cos(rad2deg(90-lat2));
    double a = acos((x1*x2 + y1*y2 + z1*z2)/pow(calcR((double) (lat1+lat2)/2),2));
    return calcR((double) (lat1+lat2) / 2) * a;
}

//#ifdef 0
// I'm still working on this...

// Lifted from gpsdrive 1.7
// CalcR gets the radius of the earth at a particular latitude
// calcxy finds the x and y positions on a 1280x1024 image of a certian scale
//  centered on a given lat/lon.

// This pulls the "real radius" of a lat, instead of a global guesstimate
double calcR (double lat)
{
    double a = 6378.137, r, sc, x, y, z;
    double e2 = 0.081082 * 0.081082;
    /*
     the radius of curvature of an ellipsoidal Earth in the plane of the
     meridian is given by

     R' = a * (1 - e^2) / (1 - e^2 * (sin(lat))^2)^(3/2)

     where a is the equatorial radius,
     b is the polar radius, and
     e is the eccentricity of the ellipsoid = sqrt(1 - b^2/a^2)

     a = 6378 km (3963 mi) Equatorial radius (surface to center distance)
     b = 6356.752 km (3950 mi) Polar radius (surface to center distance)
     e = 0.081082 Eccentricity
     */

    lat = lat * M_PI / 180.0;
    sc = sin (lat);
    x = a * (1.0 - e2);
    z = 1.0 - e2 * sc * sc;
    y = pow (z, 1.5);
    r = x / y;

    r = r * 1000.0;
    return r;
}

/* Mapscale / pixelfact is meter / pixel */
#define PIXELFACT 2817.947378

void calcxy (double *posx, double *posy, double lat, double lon, double pixelfact,
        double zero_lat, double zero_long)
{
    double dif;

    *posx = (calcR(lat) * M_PI / 180.0) * cos (M_PI * lat / 180.0) * (lon - zero_long);

    *posx = (map_width/2) + *posx / pixelfact;
    //*posx = *posx - xoff;

    *posy = (calcR(lat) * M_PI / 180.0) * (lat - zero_lat);

    dif = calcR(lat) * (1 - (cos ((M_PI * (lon - zero_long)) / 180.0)));

    *posy = *posy + dif / 1.85;
    *posy = (map_height/2) - *posy / pixelfact;
    //*posy = *posy - yoff;
}
//#endif

// Find the best map scale for the 'rectangle' tlat,tlon blat,blon
long int BestMapScale(double tlat, double tlon, double blat, double blon) {
    for (int x = 0; scales[x] != 0; x++) {
        /*
        if (((double) (scales[x] / PIXELFACT) * map_width > width) &&
            ((double) (scales[x] / PIXELFACT) * map_height > height))
            return scales[x];
            */

        double mapx, mapy;
        calcxy (&mapx, &mapy, global_map_avg.max_lat, global_map_avg.max_lon,
                (double) scales[x]/PIXELFACT, map_avg_lat, map_avg_lon);

        if (mapx < 0 || mapx > map_width || mapy < 0 || mapy > map_height)
            continue;
        else {
            // Fudge the scale by 10% for extreme ranges
            if (scales[x] >= 1000000 && scales[x] < 20000000)
                return (long) (scales[x] + (scales[x] * 0.10));
            if (scales[x] >= 20000000)
                return (long) (scales[x] + (scales[x] * 0.15));

            return scales[x];
        }
    }

    return 0;
}

#define geom_distance(a, b, x, y) sqrt(pow((double) (a) - (double) (x), 2) + pow((double) (b) - (double) (y), 2))

// Frank and Nielson's improved weighting algorithm
double WeightAlgo(int start_x, int start_y, int in_x, int in_y, double in_fuzz, double in_scale) {

    // Step 1:  Find 'r', the distance to the farthest interpolation point
    int min_x = 0, min_y = 0;
    int max_x = map_width, max_y = map_height;
    int offset = (int) (in_fuzz/in_scale);

    if (start_x - offset > min_x)
        min_x = start_x - offset;
    if (start_y - offset > min_y)
        min_y = start_y - offset;
    if (start_x + offset < max_x)
        max_x = start_x + offset;
    if (start_y + offset < max_y)
        max_y = start_y + offset;

    // Find the farthest sample point in this set
    double r = 0;
    for (int cury = min_y; cury < max_y; cury++) {
        for (int curx = min_x; curx < max_x; curx++) {
            if (power_input_map[(map_width * cury) + curx] < 0)
                continue;

            double h = geom_distance(start_x, start_y, curx, cury);
            if (h > r)
                r = h;
        }
    }

    // Construct the 'top half' of the weight function:
    double hi = geom_distance(start_x, start_y, in_x, in_y);
    double top_func = pow( ((r - hi)/(r * hi)), 2);

    double bot_sum = 0;
    // Construct the 'bottom half' of the weight function
    for (int cury = min_y; cury < max_y; cury++) {
        for (int curx = min_x; curx < max_x; curx++) {
            if (power_input_map[(map_width * cury) + curx] < 0)
                continue;

            double hj = geom_distance(start_x, start_y, curx, cury) * 1.8;

            bot_sum += pow( ((r - hj)/(r * hj)), 2);
        }
    }

    // Put it all together and return the influence
    double weight = top_func/bot_sum;

    return weight;
}


// Inverse weight calculations -- Shepard's with Frank and Nielson's improved weight
// algorithm
int InverseWeight(int in_x, int in_y, int in_fuzz, double in_scale) {

//    fprintf(stderr, "Getting power on %d %d\n", in_x, in_y);


    int min_x = 0, min_y = 0;
    int max_x = map_width, max_y = map_height;
    int offset = (int) (in_fuzz/in_scale);

    if (in_x - offset > min_x)
        min_x = in_x - offset;
    if (in_y - offset > min_y)
        min_y = in_y - offset;
    if (in_x + offset < max_x)
        max_x = in_x + offset;
    if (in_y + offset < max_y)
        max_y = in_y + offset;

    /*
     fprintf(stderr, "influenced by %d range, %d %d from %d %d to %d %d\n",
            offset, in_x, in_y, min_x, min_y, max_x, max_y);
            */

    if (offset == 0)
        return 0;

    double power_sum = 0;

    for (int cury = min_y; cury < max_y; cury++) {
        for (int curx = min_x; curx < max_x; curx++) {

            if (power_input_map[(map_width * cury) + curx] < 0)
                continue;

            power_sum += WeightAlgo(in_x, in_y, curx, cury, in_fuzz, in_scale) *
                power_input_map[(map_width * cury) + curx];

        }
    }

    return (int) power_sum;

}


int ProcessGPSFile(char *in_fname) {
    vector<gps_network> net_vec;
#ifdef HAVE_LIBZ
    gzFile gpsfz;
#else
     FILE *gpsf;
#endif

#ifdef HAVE_LIBZ
    if ((gpsfz = gzopen(in_fname, "rb")) == NULL) {
#else
    if ((gpsf = fopen(in_fname, "rb")) == NULL) {
#endif
        fprintf(stderr, "FATAL:  Could not open data file.\n");
        return -1;
    }

    file_hdr filhdr;

    /*
    if (fread(&filhdr, sizeof(file_hdr), 1, gpsf) < 1) {
        fprintf(stderr, "FATAL:  Could not read header.\n");
        return -1;
        }
        */
#ifdef HAVE_LIBZ
    if (gzread(gpsfz, &filhdr, sizeof(file_hdr)) < (int) sizeof(file_hdr))
#else
    if (fread(&filhdr, sizeof(file_hdr), 1, gpsf) < 1)
#endif
    {
        fprintf(stderr, "FATAL:  Could not read header.\n");
        return -1;
    }

    if (filhdr.magic != GPS_MAGIC) {
        fprintf(stderr, "FATAL:  Invalid gpsdump magic on %s.\n", in_fname);
        return -1;
    }

    printf("%s header version %d\n", in_fname, filhdr.version);
    if (filhdr.version > 4) {
        fprintf(stderr, "FATAL:  Unsupported version %d.\n", filhdr.version);
        return -1;
    }

    net_hdr nethdr[HEADER_NUM];
    memset(nethdr, 0, sizeof(net_hdr) * HEADER_NUM);

    // Yes this is a nasty repetition of code.
    // No I don't have a better way of doing it.
    if (filhdr.version == 1) {
        net_hdr_v1 nethdr_v1[254];
#ifdef HAVE_LIBZ
        if (gzread(gpsfz, &nethdr_v1, sizeof(net_hdr_v1) * 254) < (int) sizeof(net_hdr_v1) * 254)
#else
        if (fread(&nethdr_v1, sizeof(net_hdr_v1), 254, gpsf) < 254)
#endif
        {
            fprintf(stderr, "FATAL:  Could not read v1 network headers.\n");
            return -1;
        }

        // These are easy since they didn't change but we'll read them seperately anyhow
        for (int x = 0; x < 254; x++) {
            /*
            typedef struct {
                uint8_t number;
                uint8_t bssid[6];
                uint8_t ssid[SSID_SIZE];
                } net_hdr;
                */

            nethdr[x].number = nethdr_v1[x].number;
            memcpy(nethdr[x].bssid, nethdr_v1[x].bssid, 6);
            memcpy(nethdr[x].ssid, nethdr_v1[x].ssid, SSID_SIZE);
        }

    } else if (filhdr.version == 2) {
        net_hdr_v2 nethdr_v2[255];
#ifdef HAVE_LIBZ
        if (gzread(gpsfz, &nethdr_v2, sizeof(net_hdr_v2) * 254) < (int) sizeof(net_hdr_v2) * 254)
#else
        if (fread(&nethdr_v2, sizeof(net_hdr_v2), 254, gpsf) < 254)
#endif
        {
            fprintf(stderr, "FATAL:  Could not read v2 network headers.\n");
            return -1;
        }

        // These are easy since they didn't change but we'll read them seperately anyhow
        for (int x = 0; x < 254; x++) {
            nethdr[x].number = nethdr_v2[x].number;
            memcpy(nethdr[x].bssid, nethdr_v2[x].bssid, 6);
            memcpy(nethdr[x].ssid, nethdr_v2[x].ssid, SSID_SIZE);
        }
    } else if (filhdr.version == 3) {
        net_hdr_v3 nethdr_v3[254];
#ifdef HAVE_LIBZ
        if (gzread(gpsfz, &nethdr_v3, sizeof(net_hdr_v3) * 254) < (int) sizeof(net_hdr_v3) * 254)
#else
        if (fread(&nethdr_v3, sizeof(net_hdr_v3), 254, gpsf) < 254)
#endif
        {
            fprintf(stderr, "FATAL:  Could not read v3 network headers.\n");
            return -1;
        }

        // These are easy since they didn't change but we'll read them seperately anyhow
        for (int x = 0; x < 254; x++) {
            nethdr[x].number = nethdr_v3[x].number;
            memcpy(nethdr[x].bssid, nethdr_v3[x].bssid, 6);
            memcpy(nethdr[x].ssid, nethdr_v3[x].ssid, SSID_SIZE);
        }

    } else {
#ifdef HAVE_LIBZ
        if (gzread(gpsfz, &nethdr, sizeof(net_hdr) * HEADER_NUM) < (int) sizeof(net_hdr) * HEADER_NUM)
#else
        if (fread(&nethdr, sizeof(net_hdr), HEADER_NUM, gpsf) < HEADER_NUM)
#endif
        {
            fprintf(stderr, "FATAL:  Could not read v4 network headers.\n");
            return -1;
        }
    }


    // Fill the network vector with empty networks
    for (int x = 0; x < HEADER_NUM; x++) {
        gps_network netw;

//        printf("Got header '%s'\n", nethdr[x].ssid);

        memset(&netw, 0, sizeof(gps_network));

        netw.header = nethdr[x];

        /*
        printf("got network header %d encode number %d '%s'\n",
        x, netw.header.number, netw.header.ssid);
        */

        net_vec.push_back(netw);
    }

    data_pkt dpkt;
    data_pkt_v1 dpkt_v1;
    data_pkt_v2 dpkt_v2;

    // Build a new track
    vector<track_data> trak;
    track_vec.push_back(trak);
    num_tracks++;

    int last_power = 0;
    int power_count = 0;

    // Read all the data packets
    /*
     while (fread(&dpkt, sizeof(dpkt), 1, gpsf) > 0) {
     */

    int read_ret = 1;

    int file_samples = 0;
    while (read_ret > 0) {

        file_samples++;

        if (filhdr.version == 1) {
#ifdef HAVE_LIBZ
            read_ret = gzread(gpsfz, &dpkt_v1, sizeof(dpkt_v1));
            if (read_ret < (int) sizeof(dpkt_v1))
                break;

#else
            read_ret = fread(&dpkt_v1, sizeof(dpkt_v1), 1, gpsf);
            if (read_ret < 1)
                break;
#endif

            dpkt.number = dpkt_v1.number;
            dpkt.lat = dpkt_v1.lat;
            dpkt.lat_mant = dpkt_v1.lat_mant;
            dpkt.lon = dpkt_v1.lon;
            dpkt.lon_mant = dpkt_v1.lon_mant;
            dpkt.alt = dpkt_v1.alt;
            dpkt.alt_mant = dpkt_v1.alt_mant;
            dpkt.spd = dpkt_v1.spd;
            dpkt.spd_mant = dpkt_v1.spd_mant;
            dpkt.fix = dpkt_v1.fix;
            dpkt.power = 0;
            memcpy(&dpkt.ts, &dpkt_v1.ts, sizeof(time_hdr));
        } else if (filhdr.version == 2 || filhdr.version == 3) {
#ifdef HAVE_LIBZ
            read_ret = gzread(gpsfz, &dpkt_v2, sizeof(dpkt_v2));
            if (read_ret < (int) sizeof(dpkt_v2))
                break;
#else
            read_ret = fread(&dpkt_v2, sizeof(dpkt_v2), 1, gpsf);
            if (read_ret < 1)
                break;
#endif

            dpkt.number = dpkt_v2.number;
            dpkt.lat = dpkt_v2.lat;
            dpkt.lat_mant = dpkt_v2.lat_mant;
            dpkt.lon = dpkt_v2.lon;
            dpkt.lon_mant = dpkt_v2.lon_mant;
            dpkt.alt = dpkt_v2.alt;
            dpkt.alt_mant = dpkt_v2.alt_mant;
            dpkt.spd = dpkt_v2.spd;
            dpkt.spd_mant = dpkt_v2.spd_mant;
            dpkt.fix = dpkt_v2.fix;
            dpkt.power = dpkt_v2.power;
            memcpy(&dpkt.ts, &dpkt_v1.ts, sizeof(time_hdr));

        } else {
#ifdef HAVE_LIBZ
            read_ret = gzread(gpsfz, &dpkt, sizeof(dpkt));

            if (read_ret < (int) sizeof(dpkt))
                break;

#else
            read_ret = fread(&dpkt, sizeof(dpkt), 1, gpsf);

            if (read_ret < 1)
                break;
#endif
        }

        // We can't trust packets that have no coordinate fix
        if (dpkt.fix == 0)
            continue;

        double lat, lon, alt, spd;

        // Convert the stored primary+mantissa into a double
        lat = (double) dpkt.lat + ((double) dpkt.lat_mant / 1000000);
        lon = (double) dpkt.lon + ((double) dpkt.lon_mant / 1000000);
        alt = (double) dpkt.alt + ((double) dpkt.alt_mant / 1000000);
        spd = (double) dpkt.spd + ((double) dpkt.spd_mant / 1000000);

        if (lat == 0 || lon == 0)
            continue;

        /*
        if (Mac2String(net_vec[dpkt.number].header.bssid, ':') == "00:04:5A:0E:0E:62")
            printf("WEIRD:  Got %3.6fx%3.6f fix %d for weirdo network in %s.\n",
                   lat, lon, dpkt.fix, in_fname);
                   */

        // Add to the global average -- NOTE -- This will need to check for what
        // SSIDs to use to scale correctly
        global_map_avg.avg_lat += lat;
        global_map_avg.avg_lon += lon;
        global_map_avg.avg_alt += alt;
        global_map_avg.avg_spd += spd;
        global_map_avg.count++;

        // Don't calculate track data for map size if we're not drawing tracks.
        if (!(draw_track == 0 &&
              ((dpkt.number == 0xFF && filhdr.version <= 3) ||
               (dpkt.number == HEADER_NUM && filhdr.version > 3)) )) {
            if (lat > global_map_avg.max_lat || global_map_avg.max_lat == 0)
                global_map_avg.max_lat = lat;
            if (lat < global_map_avg.min_lat || global_map_avg.min_lat == 0)
                global_map_avg.min_lat = lat;

            if (lon > global_map_avg.max_lon || global_map_avg.max_lon == 0)
                global_map_avg.max_lon = lon;
            if (lon < global_map_avg.min_lon || global_map_avg.min_lon == 0)
                global_map_avg.min_lon = lon;

            if (alt > global_map_avg.max_alt || global_map_avg.max_alt == 0)
                global_map_avg.max_alt = alt;
            if (alt < global_map_avg.min_alt || global_map_avg.min_alt == 0)
                global_map_avg.min_alt = alt;
        }

        // Push the path into the current track vector
        if ((dpkt.number == 0xFF && filhdr.version <= 3) ||
            (dpkt.number == HEADER_NUM && filhdr.version > 3)) {
            track_data tdat;

            // Junk unfixed packets
            if (dpkt.fix < 2)
                continue;

            tdat.x = 0;
            tdat.y = 0;

            tdat.lat = lat;
            tdat.lon = lon;

            tdat.version = filhdr.version;

            if (dpkt.power == last_power) {
                if (power_count < 3)
                    tdat.power = dpkt.power;
                else
                    tdat.power = 0;
                power_count++;
            } else {
                last_power = dpkt.power;
                power_count = 0;
                tdat.power = dpkt.power;
            }

            if (tdat.power != 0)
                power_data = 1;

            track_vec[num_tracks-1].push_back(tdat);
        }

        // Enter it into our current vector
        net_vec[dpkt.number].avg_lat += lat;
        net_vec[dpkt.number].avg_lon += lon;
        net_vec[dpkt.number].avg_alt += alt;
        net_vec[dpkt.number].avg_spd += spd;
        net_vec[dpkt.number].count++;

        // Enter the max/min values
        if (lat > net_vec[dpkt.number].max_lat || net_vec[dpkt.number].max_lat == 0)
            net_vec[dpkt.number].max_lat = lat;

        if (lat < net_vec[dpkt.number].min_lat || net_vec[dpkt.number].min_lat == 0)
            net_vec[dpkt.number].min_lat = lat;

        if (lon > net_vec[dpkt.number].max_lon || net_vec[dpkt.number].max_lon == 0)
            net_vec[dpkt.number].max_lon = lon;

        if (lon < net_vec[dpkt.number].min_lon || net_vec[dpkt.number].min_lon == 0)
            net_vec[dpkt.number].min_lon = lon;

        if (alt > net_vec[dpkt.number].max_alt || net_vec[dpkt.number].max_alt == 0)
            net_vec[dpkt.number].max_alt = alt;

        if (alt < net_vec[dpkt.number].min_alt || net_vec[dpkt.number].min_alt == 0)
            net_vec[dpkt.number].min_alt = alt;

    }

    // Now transform our per-file list of 254 networks into the global BSSID records
    for (unsigned int x = 0; x < net_vec.size(); x++) {
        gps_network vec_iter = net_vec[x];
        string bssid = Mac2String(vec_iter.header.bssid, ':');

        // Skip data only networks
        if (bssid == "00:00:00:00:00:00" || bssid == "FF:FF:FF:FF:FF:FF")
            continue;

        if (vec_iter.count <= 0)
            continue;

        // Make a new record or add to the existing one
        if (net_map.find(bssid) == net_map.end()) {
            net_map[bssid] = vec_iter;
        } else {
            gps_network map_iter = net_map[bssid];

            // Add the average data
            map_iter.count += vec_iter.count;
            map_iter.avg_lat += vec_iter.avg_lat;
            map_iter.avg_lon += vec_iter.avg_lon;
            map_iter.avg_alt += vec_iter.avg_alt;
            map_iter.avg_spd += vec_iter.avg_spd;

            // Calc our new min and max
            if (vec_iter.max_lat > map_iter.max_lat || map_iter.max_lat == 0)
                map_iter.max_lat = vec_iter.max_lat;
            if (vec_iter.min_lat < map_iter.min_lat || map_iter.min_lat == 0)
                map_iter.min_lat = vec_iter.min_lat;
            if (vec_iter.max_lon > map_iter.max_lon || map_iter.max_lon == 0)
                map_iter.max_lon = vec_iter.max_lon;
            if (vec_iter.min_lon < map_iter.min_lon || map_iter.min_lon == 0)
                map_iter.min_lon = vec_iter.min_lon;
            if (vec_iter.max_alt > map_iter.max_alt || map_iter.max_alt == 0)
                map_iter.max_alt = vec_iter.max_alt;
            if (vec_iter.min_alt < map_iter.min_alt || map_iter.min_alt == 0)
                map_iter.min_alt = vec_iter.min_alt;

            // And put it back in
            net_map[bssid] = map_iter;
        }
    }

    /*
     fclose(gpsf);
     */
#ifdef HAVE_LIBZ
    gzclose(gpsfz);
#else
     fclose(gpsf);
#endif

     if (verbose)
         fprintf(stderr, "%s contains %d samples.\n",
                 in_fname, file_samples);

     sample_points += file_samples;

     return 1;
}

// Do all the math
void ProcessNetData(int in_printstats) {

    // Convert the tracks to x,y
    for (unsigned int vec = 0; vec < track_vec.size(); vec++) {
        for (unsigned int x = 0; x < track_vec[vec].size(); x++) {
            double track_tx, track_ty;
            calcxy(&track_tx, &track_ty, track_vec[vec][x].lat, track_vec[vec][x].lon,
                   (double) scale/PIXELFACT, map_avg_lat, map_avg_lon);

            /*
            if (draw_track == 1 && (track_ty < 0 || track_ty > map_height || track_tx < 0 || track_tx > map_width)) {
                //printf("Out of bounds, %f,%f %f,%f fix\n", track_tx, track_ty, track_vec[vec][x].lat, track_vec[vec][x].lon);
                continue;
                }
                */

            track_vec[vec][x].x = (int) track_tx;
            track_vec[vec][x].y = (int) track_ty;
        }

        if (in_printstats)
            printf("Track %d: %d samples.\n", vec, track_vec[vec].size());
    }

    printf("Graphing %d networks.\n", net_map.size());

    for (map<string, gps_network>::const_iterator x = net_map.begin();
         x != net_map.end(); ++x) {

        gps_network map_iter = x->second;

        if (map_iter.count <= 1)
            continue;

        map_iter.diagonal_distance = earth_distance(map_iter.max_lat, map_iter.max_lon,
                                                    map_iter.min_lat, map_iter.min_lon);

        map_iter.altitude_distance = map_iter.max_alt - map_iter.min_alt;

        double avg_lat = (double) map_iter.avg_lat / map_iter.count;
        double avg_lon = (double) map_iter.avg_lon / map_iter.count;
        double avg_alt = (double) map_iter.avg_alt / map_iter.count;
        double avg_spd = (double) map_iter.avg_spd / map_iter.count;

        map_iter.avg_lat = avg_lat;
        map_iter.avg_lon = avg_lon;
        map_iter.avg_alt = avg_alt;
        map_iter.avg_spd = avg_spd;

        net_map[Mac2String(map_iter.header.bssid, ':')] = map_iter;

        if (in_printstats)
            printf("Net:     %s [%s]\n"
                   "  Samples : %d\n"
                   "  Min lat : %f\n"
                   "  Min lon : %f\n"
                   "  Max lat : %f\n"
                   "  Max lon : %f\n"
                   "  Min alt : %f\n"
                   "  Max Alt : %f\n"
                   "  Avg Lat : %f\n"
                   "  Avg Lon : %f\n"
                   "  Avg Alt : %f\n"
                   "  Avg Spd : %f\n"
                   "  H. Range: %f ft\n"
                   "  V. Range: %f ft\n",
                   map_iter.header.ssid,
                   Mac2String(map_iter.header.bssid, ':').c_str(),
                   map_iter.count,
                   map_iter.min_lat, map_iter.min_lon,
                   map_iter.max_lat, map_iter.max_lon,
                   map_iter.min_alt, map_iter.max_alt,
                   map_iter.avg_lat, map_iter.avg_lon,
                   map_iter.avg_alt, map_iter.avg_spd,
                   map_iter.diagonal_distance * 3.3, map_iter.altitude_distance);
    }
}

void AssignNetColors() {
    int base_color = 1;

    for (map<string, gps_network>::const_iterator x = net_map.begin();
         x != net_map.end(); ++x) {

        gps_network map_iter = x->second;

        if (netcolors[base_color] == NULL)
            base_color = 1;

        map_iter.color_index = base_color;

        net_map[Mac2String(map_iter.header.bssid, ':')] = map_iter;

        base_color++;
    }
}

void DrawNetTracks(Image *in_img, DrawInfo *in_di) {
    // Our track color
    uint8_t track_r = 0x00, track_g = 0x00, track_b = 0xFF;
    char color_str[8];
    PixelPacket track_clr;

    // Draw each track
    for (unsigned int vec = 0; vec < track_vec.size(); vec++) {
        if (track_vec[vec].size() == 0)
            continue;

        // Generate the color we're drawing with
        snprintf(color_str, 8, "#%02X%02X%02X", track_r, track_g, track_b);
        
#ifdef OldMagickLib
	QueryColorDatabase(color_str, &track_clr);
#else
	{
	ExceptionInfo excep;
	QueryColorDatabase(color_str, &track_clr, &excep);
	}
#endif

        in_di->stroke = track_clr;

        // Dim the color
        track_b -= track_decay;

        // Reset it if we're "too dark"
        if (track_b < 0x50)
            track_b = 0xFF;

        // Initialize the previous track location vars

        /*
        double prev_tlat = 0, prev_tlon = 0;
        prev_tlat = track_vec[vec][0].lat;
        prev_tlon = track_vec[vec][0].lon;


        double prev_tx, prev_ty;
        calcxy(&prev_tx, &prev_ty, prev_tlat, prev_tlon,
        (double) scale/PIXELFACT, map_avg_lat, map_avg_lon);
        */

        int prev_tx, prev_ty;
        prev_tx = track_vec[vec][0].x;
        prev_ty = track_vec[vec][0].y;

        for (unsigned int x = 1; x < track_vec[vec].size(); x++) {
            char prim[1024];

            // If we don't have a previous vector (ie, the map data failed), set it
            // and continue
            if (prev_tx == -1 || prev_ty == -1) {
                prev_tx = track_vec[vec][x].x;
                prev_ty = track_vec[vec][x].y;
                continue;
            }

            // Scrap dupes
            if (track_vec[vec][x].x == (unsigned int) prev_tx &&
                track_vec[vec][x].y == (unsigned int) prev_ty)
                continue;

            // Scrap stuff entirely off-screen to save on speed
            if (((unsigned int) prev_tx > map_width && (unsigned int) prev_ty > map_height &&
                 track_vec[vec][x].x > map_width && track_vec[vec][x].y > map_height) ||
                (prev_tx < 0 && prev_ty < 0 &&
                 track_vec[vec][x].x < 0 && track_vec[vec][x].y < 0))
                continue;

            // If the track jumps more than 50 meters in 1 second, assume we had a
            // problem and restart the track at the next position
            double distance;
            if ((distance = geom_distance(track_vec[vec][x].x, track_vec[vec][x].y,
                                          prev_tx, prev_ty)) * (scale/PIXELFACT) > 50) {
                prev_tx = -1;
                prev_ty = -1;
                continue;
            }

            /* Don't whine about track jumps (for now)
            if (sqrt(pow(track_vec[vec][x].x - prev_tx, 2) + pow(track_vec[vec][x].y - prev_ty, 2)) > 20) {
                printf("Suspicious track record: %dx%d (%fx%f)\n"
                       "Prev: %dx%d (%fx%f)\n",
                       track_vec[vec][x].x, track_vec[vec][x].y,
                       track_vec[vec][x].lat, track_vec[vec][x].lon,
                       prev_tx, prev_ty,
                       track_vec[vec][x-1].lat, track_vec[vec][x-1].lon);
                       }
                       */

            /*
            double track_tx, track_ty;
            calcxy(&track_tx, &track_ty, track_vec[vec][x].lat, track_vec[vec][x].lon,
            (double) scale/PIXELFACT, map_avg_lat, map_avg_lon);
            */

            snprintf(prim, 1024, "stroke-width %d line %d,%d %d,%d",
                     track_width,
                     prev_tx, prev_ty, track_vec[vec][x].x, track_vec[vec][x].y);

                     /*
                     (int) prev_tx, (int) prev_ty,
                     (int) track_tx, (int) track_ty);
                     */

            in_di->primitive = strdup(prim);
            DrawImage(in_img, in_di);

            /*
            prev_tlat = track_vec[vec][x].lat;
            prev_tlon = track_vec[vec][x].lon;

            prev_tx = track_tx;
            prev_ty = track_ty;
            */

            prev_tx = track_vec[vec][x].x;
            prev_ty = track_vec[vec][x].y;
        }
    }
}

void DrawNetCircles(Image *in_img, DrawInfo *in_di) {
    for (map<string, gps_network>::const_iterator x = net_map.begin();
         x != net_map.end(); ++x) {

        gps_network map_iter = x->second;

        // Skip networks w/ no determined coordinates
        if (map_iter.max_lat == 0)
            continue;

        if (map_iter.diagonal_distance > horiz_throttle)
            continue;

        // Figure x, y of min on our hypothetical map
        double mapx, mapy;

        calcxy (&mapx, &mapy, map_iter.avg_lat, map_iter.avg_lon,
                (double) scale/PIXELFACT, map_avg_lat, map_avg_lon);

        /*
        printf("  Pixel factoring:  %f\n", (double) 25000/PIXELFACT);
        printf("  Radius:  %f\n", calcR(net_vec[x].min_lat));
        printf("  Coords  : %dx%d\n", (int) mapx, (int) mapy);
        */

        double end_lat, end_lon;

        // Find the nearest corner of the bounding rectangle, this will determine
        // the size of our network circle
        if (((map_iter.min_lat + map_iter.max_lat) / 2) < map_iter.avg_lat)
            end_lat = map_iter.max_lat;
        else
            end_lat = map_iter.min_lat;
        if (((map_iter.min_lon + map_iter.max_lon)/ 2) < map_iter.avg_lon)
            end_lon = map_iter.max_lon;
        else
            end_lon = map_iter.min_lon;

        /*
        if ((map_iter.min_lat - avg_lat) < (map_iter.max_lat - avg_lat) &&
            (map_iter.min_lon - avg_lon) < (map_iter.max_lon - avg_lon)) {
            end_lat = map_iter.min_lat;
            end_lon = map_iter.min_lon;
        } else {
            end_lat = map_iter.max_lat;
            end_lon = map_iter.min_lon;
            }
            */

        double endx, endy;
        calcxy(&endx, &endy, end_lat, end_lon,
               (double) scale/PIXELFACT, map_avg_lat, map_avg_lon);

        // printf("  Endpt   : %dx%d\n", (int) endx, (int) endy);

        PixelPacket netclr;

#ifdef OldMagickLib
	QueryColorDatabase(netcolors[map_iter.color_index], &netclr);
#else
	{
	ExceptionInfo excep;
 	QueryColorDatabase(netcolors[map_iter.color_index], &netclr, &excep);
	}
#endif

        in_di->fill = netclr;

        char prim[1024];

        snprintf(prim, 1024, "fill-opacity %d%% stroke-opacity %d%% circle %d,%d %d,%d",
                 draw_opacity, draw_opacity, (int) mapx, (int) mapy, (int) endx, (int) endy);

                /*

        printf("%s [%s] %s\n", map_iter.header.ssid, Mac2String(map_iter.header.bssid, ':').c_str(),
               prim);

        if (Mac2String(map_iter.header.bssid, ':') == "00:40:96:23:B1:EB") {
            printf("Flukey network:\n");

            printf("color: %d\n", map_iter.color_index);

            snprintf(prim, 1024, "circle %d,%d %d,%d",
                     (int) mapx, (int) mapy, (int) endx, (int) endy);

                     }

                     printf("%s\n", prim);
                     */


        in_di->primitive = strdup(prim);
        DrawImage(in_img, in_di);

        /*
         printf("Drew circle.\n");
         */

        /*
        di->fill = black;
        snprintf(prim, 1024, "text %d,%d \"%s %s\"",
                 (int) mapx, (int) mapy,
                 net_vec[x].header.ssid,
                 Mac2String(net_vec[x].header.bssid, ':').c_str());
        di->primitive = strdup(prim);
        DrawImage(img, di);
        */

        /*
        //d = sqrt[(x1-x2)^2 + (y1-y2)^2]
        printf("  Px RLen : %d\n", (int) sqrt(pow((int) mapx - endx, 2) + pow((int) mapy - endy, 2)));
        */

    }
}

void DrawNetBoundRects(Image *in_img, DrawInfo *in_di, int in_fill) {
    for (map<string, gps_network>::const_iterator x = net_map.begin();
         x != net_map.end(); ++x) {

        gps_network map_iter = x->second;

        // Skip networks w/ no determined coordinates
        if (map_iter.max_lat == 0)
            continue;

        if (map_iter.diagonal_distance == 0 || map_iter.diagonal_distance > horiz_throttle)
            continue;

        // Figure x, y of min on our hypothetical map
        double mapx, mapy;
        calcxy (&mapx, &mapy, map_iter.max_lat, map_iter.max_lon,
                (double) scale/PIXELFACT, map_avg_lat, map_avg_lon);

        double endx, endy;
        calcxy(&endx, &endy, map_iter.min_lat, map_iter.min_lon,
               (double) scale/PIXELFACT, map_avg_lat, map_avg_lon);

        double tlx, tly, brx, bry;

        if (mapx < endx) {
            tlx = mapx;
            brx = endx;
        } else {
            tlx = endx;
            brx = mapx;
        }

        if (mapy < endy) {
            tly = mapy;
            bry = endy;
        } else {
            tly = endy;
            bry = mapy;
        }

        if (in_fill) {
            PixelPacket netclr;

#ifdef OldMagickLib
	QueryColorDatabase(netcolors[map_iter.color_index], &netclr);
#else
	{
	ExceptionInfo excep;
	QueryColorDatabase(netcolors[map_iter.color_index], &netclr, &excep);
	}
#endif
            
            in_di->fill = netclr;
        }

        char prim[1024];

        snprintf(prim, 1024, "stroke black fill-opacity %d%% rectangle %d,%d %d,%d",
                 in_fill, (int) mapx, (int) mapy, (int) endx, (int) endy);

        //snprintf(prim, 1024, "fill-opacity %d%% rectangle %d,%d %d,%d",
        //in_fill, (int) tlx, (int) tly, (int) brx, (int) bry);


        in_di->primitive = strdup(prim);
        DrawImage(in_img, in_di);

        /*
        //d = sqrt[(x1-x2)^2 + (y1-y2)^2]
        printf("  Px RLen : %d\n", (int) sqrt(pow((int) mapx - endx, 2) + pow((int) mapy - endy, 2)));
        */

    }
}

void DrawLegend(Image *in_img, DrawInfo *in_di) {



}

// Thread function to compute a line of interpolated data

typedef struct powerline_arg {
//    unsigned int y;
//    unsigned int y_max;
    unsigned int in_res;
    unsigned int threadno;
};

void *PowerLine(void *arg) {
    powerline_arg *parg = (powerline_arg *) arg;
    time_t startline;

//    unsigned int y_offset = parg->y;
//    unsigned int y_max = parg->y_max;
    unsigned int in_res = parg->in_res;
    unsigned int y = 0;

    while (y < map_height) {
        pthread_mutex_lock(&power_pos_lock);
        y = power_pos * in_res;
        power_pos++;
        pthread_mutex_unlock(&power_pos_lock);

        if (y >= map_height)
            break;

        //    for (unsigned int y = y_offset; y < map_height; y+= (in_res * numthreads)) {
        startline = time(0);

        pthread_mutex_lock(&print_lock);
        fprintf(stderr, "Thread %d: crunching interpolation image line %d\n", parg->threadno, y);
        pthread_mutex_unlock(&print_lock);

        for (unsigned int x = 0; x < map_width; x+= in_res) {
            unsigned int powr = InverseWeight(x, y, (int) (scale/100), (double) scale/PIXELFACT);

            pthread_mutex_lock(&power_lock);
            power_map[(map_width * y) + x] = powr;
            pthread_mutex_unlock(&power_lock);
        }

        if (verbose) {
            pthread_mutex_lock(&print_lock);
            int elapsed = time(0) - startline;
            int complet = elapsed * ((map_height - y) / in_res);

            fprintf(stderr, "Completed in %d seconds.  (Estimated: %dh %dm %ds to completion)\n",
                    elapsed, (complet/60)/60, (complet/60) % 60, complet % 60);
            pthread_mutex_unlock(&print_lock);
        }

    }

    pthread_exit((void *) 0);
}

void DrawNetPower(Image *in_img, DrawInfo *in_di, int in_res) {
    PixelPacket point_clr;
    pthread_attr_t attr;

    power_map = (unsigned int *) malloc(sizeof(unsigned int) * (map_width * map_height));
    memset(power_map, 0, sizeof(unsigned int) * (map_width * map_height));

    power_input_map = (int *) malloc(sizeof(int) * (map_width * map_height));
    memset(power_input_map, -1, sizeof(int) * (map_width * map_height));

    // Convert the power data in the tracks into a 2d map
    fprintf(stderr, "Converting track power data to 2d graph...\n");
    for (unsigned int vec = 0; vec < track_vec.size(); vec++) {
        for(unsigned int i = 0; i < track_vec[vec].size(); i++) {

            if (track_vec[vec][i].version < 2)
                continue;

            unsigned int curx = track_vec[vec][i].x, cury = track_vec[vec][i].y;

            if (curx >= map_width || cury >= map_height || curx < 0 || cury < 0)
                continue;

            if (power_input_map[(map_width * cury) + curx] < track_vec[vec][i].power)
                power_input_map[(map_width * cury) + curx] = track_vec[vec][i].power;

        }
    }

    fprintf(stderr, "Interpolating power into graph points.\n");
    // Slice the map into pieces and assign it to the threads, averaging high - if it's
    // not evenly divisible the last thread may get less work to do than the others.
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
    powerline_arg *pargs = new powerline_arg[numthreads];
    for (unsigned int t = 0; t < numthreads; t++) {
        pargs[t].in_res = in_res;
        pargs[t].threadno = t;

        pthread_create(&mapthread[t], &attr, PowerLine, (void *) &pargs[t]);
    }

    pthread_attr_destroy(&attr);

    // Now wait for the threads to complete and come back
    int thread_status;
    for (unsigned int t = 0; t < numthreads; t++) {
        pthread_join(mapthread[t], (void **) &thread_status);
    }

    fprintf(stderr, "Drawing interpolated power levels to map.\n");
    for (unsigned int y = 0; y < map_height; y += in_res) {
        for (unsigned int x = 0; x < map_width; x += in_res) {
            //            int powr = InverseWeight(x, y, 200, (double) scale/PIXELFACT);
            int powr = power_map[(map_width * y) + x];

            //printf("Got weight %d for pixel %d,%d\n", powr, x, y);

            if (powr > 0) {

                int power_index = powr / (power_max / power_steps);
                if (powr < (power_max / power_steps) / 2)
                    continue;

                if (power_index >= power_steps)
                    power_index = power_steps - 1;

#ifdef OldMagickLib
		QueryColorDatabase(powercolors[power_index], &point_clr);
#else
		{
		ExceptionInfo excep;
		QueryColorDatabase(powercolors[power_index], &point_clr, &excep);
		}
#endif

                in_di->stroke = point_clr;
                in_di->fill = point_clr;

                char prim[1024];

                int b, r;

                if (in_res == 1) {
                    snprintf(prim, 1024, "fill-opacity %d%% stroke-opacity %d%% stroke-width 0 point %d,%d",
                         draw_opacity, draw_opacity, x, y);
                } else {
                    r = x + in_res - 1;
                    b = y + in_res - 1;
                    snprintf(prim, 1024, "fill-opacity %d%% stroke-opacity %d%% stroke-width 0 rectangle %d,%d %d,%d",
                         draw_opacity, draw_opacity, x, y, r, b);
                }

                //printf("%d,%d power %d\n", x, y, powr);
                //snprintf(prim, 1024, "fill-opacity %d%% stroke-opacity %d%% stroke-width 1 rectangle %d,%d %d,%d",
                //         draw_opacity, draw_opacity, x, y, r, b);

                in_di->primitive = strdup(prim);
                DrawImage(in_img, in_di);

            }
        }

    }

    delete power_map;
}

void DrawNetCenterDot(Image *in_img, DrawInfo *in_di, int in_dotsize) {
    for (map<string, gps_network>::const_iterator x = net_map.begin();
         x != net_map.end(); ++x) {

        gps_network map_iter = x->second;

        // Skip networks w/ no determined coordinates
        if (map_iter.max_lat == 0)
            continue;

        if (map_iter.diagonal_distance > horiz_throttle)
            continue;


        // Figure x, y of min on our hypothetical map
        double mapx, mapy;

        calcxy (&mapx, &mapy, map_iter.avg_lat, map_iter.avg_lon,
                (double) scale/PIXELFACT, map_avg_lat, map_avg_lon);

        double endx, endy;
        endx = mapx + in_dotsize;
        endy = mapy + in_dotsize;

        // printf("  Endpt   : %dx%d\n", (int) endx, (int) endy);

        PixelPacket netclr;
#ifdef OldMagickLib
	QueryColorDatabase(netcolors[map_iter.color_index], &netclr);
#else
	{
	ExceptionInfo excep;
	QueryColorDatabase(netcolors[map_iter.color_index], &netclr, &excep);
	}
#endif
        in_di->fill = netclr;

        char prim[1024];

        snprintf(prim, 1024, "fill-opacity 100%% stroke-opacity 100%% circle %d,%d %d,%d",
                 (int) mapx, (int) mapy, (int) endx, (int) endy);
        in_di->primitive = strdup(prim);
        DrawImage(in_img, in_di);

        /*
        di->fill = black;
        snprintf(prim, 1024, "text %d,%d \"%s %s\"",
                 (int) mapx, (int) mapy,
                 net_vec[x].heade
                 r.ssid,
                 Mac2String(net_vec[x].header.bssid, ':').c_str());
        di->primitive = strdup(prim);
        DrawImage(img, di);
        */

        /*
        //d = sqrt[(x1-x2)^2 + (y1-y2)^2]
        printf("  Px RLen : %d\n", (int) sqrt(pow((int) mapx - endx, 2) + pow((int) mapy - endy, 2)));
        */

    }
}


int Usage(char *argv) {
    printf("Usage: %s [OPTION] <GPS files>\n", argv);
    printf(
           "  -c, --coords <x,y,s>         Force map center at x,y of scale s\n"
           "  -u, --user-image-size <x,y>  Custom map size\n"
           "  -m, --user-map <map>         Use custom map instead of downloading\n"
           "  -t, --disable-draw-track     Disable drawing travel track\n"
           "  -b, --draw-bounds            Draw bounding rectangles\n"
           "  -r, --disable-draw-range     Disable drawing estimated range circles\n"
           "  -i, --draw-interpolated      Draw interpolated power (slow!)\n"
           "  -s, --interpolated-size      Resolution to interpolate with\n"
           "  -C, --interpolated-color     Alternate interpolated power color set\n"
           "                               1 is a gradiant though RGB colorspace (12 colors)\n"
           "                               2 is weathermap radar style (9 colors)\n"
           "  -d, --draw-center-dot [size] Draw a dot at the center of the network range\n"
           "  -O, --opacity <num>          Opacity of drawn elements [Default: 70]\n"
           "  -g, --no-greyscale           Don't convert map to greyscale\n"
           "  -k, --keep-gif               Keep the downloaded GIF\n"
           "  -T, --threads <num>          Number of simultaneous threads used for complex\n"
           "                               operations\n"
           "  -M, --metric                 Fetch metric-titled map\n"
           "  -v, --verbose                Verbose output while running\n"
           "  -o, --output <filename>      Output .png file\n"
           "  -v, --version                GPSMap version\n"
           "  -h, --help                   What do you think you're reading?\n");
    exit(1);
}

char *exec_name;

int main(int argc, char *argv[]) {
    exec_name = argv[0];

    char mapname[1024];
    char mapoutname[1024];

    unsigned int metric = 0;

    static struct option long_options[] = {   /* options table */
        { "disable-draw-track", no_argument, 0, 't' },
        { "draw-bounds", no_argument, 0, 'b' },
        { "disable-draw-range", no_argument, 0, 'r' },
        { "draw-interpolated", no_argument, 0, 'i' },
        { "interpolated-size", required_argument, 0, 's' },
        { "disable-draw-legend", no_argument, 0, 'l' },
        { "net-colors", required_argument, 0, 'c' },
        { "output", required_argument, 0, 'o' },
        { "opacity", required_argument, 0, 'O' },
        { "no-greyscale", no_argument, 0, 'g' },
        { "keep-gif", no_argument, 0, 'k' },
        { "user-image-size", required_argument, 0, 'u' },
        { "user-map", required_argument, 0, 'm' },
        { "quiet", no_argument, 0, 'q' },
        { "threads", required_argument, 0, 'T' },
        { "help", no_argument, 0, 'h' },
        { "version", no_argument, 0, 'v' },
        { "draw-center-dot", optional_argument, 0, 'd' },
        { "interpolated-color", required_argument, 0, 'C' },
        { "metric", no_argument, 0, 'M' },
        { 0, 0, 0, 0 }
    };
    int option_index;

    int usermap = 0;
    memset(mapname, 0, 1024);
    memset(mapoutname, 0, 1024);

    draw_track = 1;
    draw_bounds = 0;
    draw_range = 1;
    draw_interpolated = 0;
    draw_legend = 1;
    convert_greyscale = 1;
    keep_gif = 0;
    draw_center = 0;
    center_dot = 0;
    draw_opacity = 70;
    verbose = 0;
    power_data = 0;
    interpolation_res = 2;
    power_steps = power_steps_Orig;
    powercolors = powercolors_Orig;

    float user_lat, user_lon;
    long user_scale = 0;

    sample_points = 0;

    while(1) {
        int r = getopt_long(argc, argv, "T:tbrils:c:o:m:O:gqu:hvd::kC:M",
                            long_options, &option_index);

        if (r < 0) break;

        switch(r) {
        case 't':
            draw_track = 0;
            break;
        case 'b':
            draw_bounds = 1;
            break;
        case 'r':
            draw_range = 0;
            break;
        case 'i':
            draw_interpolated = 1;
            break;
        case 's':
            if (sscanf(optarg, "%d", &interpolation_res) != 1) {
                fprintf(stderr, "Invalid interpolation resolution.\n");
                Usage(argv[0]);
            }

            if (interpolation_res <= 0) {
                fprintf(stderr, "Interpolation resolution must be 1 or higher.  Setting to 2.\n");
                interpolation_res = 2;
            }

            break;
        case 'l':
            draw_legend = 0;
            break;
        case 'c':
            if (sscanf(optarg, "%f,%f,%ld", &user_lat, &user_lon,
                       &user_scale) != 3) {
                fprintf(stderr, "Invalid custom map coordinates.\n");
                Usage(argv[0]);
            }
            break;
        case 'u':
            if (sscanf(optarg, "%d,%d", &map_width, &map_height) != 2) {
                fprintf(stderr, "Invalid custom map size.\n");
                Usage(argv[0]);
            }
            break;
        case 'm':
            snprintf(mapname, 1024, "%s", optarg);
            usermap = 1;
            break;
        case 'M':
            metric = 1;
            break;
        case 'o':
            snprintf(mapoutname, 1024, "%s", optarg);
            break;
        case 'T':
            if (sscanf(optarg, "%d", &numthreads) != 1) {
                fprintf(stderr, "Invalid number of threads.\n");
                Usage(argv[0]);
            }
            break;
        case 'O':
            if (sscanf(optarg, "%d", &draw_opacity) != 1) {
                fprintf(stderr, "Invalid opacity.\n");
                Usage(argv[0]);
            }
            break;
        case 'd':
            if (optarg == NULL)
                center_dot = 3;
            else if (sscanf(optarg, "%d", &center_dot) != 1) {
                fprintf(stderr, "Invalid center dot size.\n");
                Usage(argv[0]);
            }

            draw_center = 1;
            break;
        case 'g':
            convert_greyscale = 0;
            break;
        case 'v':
            verbose = 1;
            break;
        case 'k':
            keep_gif = 1;
            break;
        case 'C':
            {
            int icolor;
            if (sscanf(optarg, "%d", &icolor) !=1) {
                fprintf(stderr, "Invalid interpolation color set\n");
                Usage(argv[0]);
            }
            if (icolor == 1) {
   		power_steps = power_steps_Math;
    		powercolors = powercolors_Math;
            } else if (icolor == 2) {
   		power_steps = power_steps_Radar;
    		powercolors = powercolors_Radar;
            }
            }
            break;
        default:
            Usage(argv[0]);
            break;
        }
    }

    // Initialize stuff
    num_tracks = 0;
    memset(&global_map_avg, 0, sizeof(gps_network));

    // Build the threads
    mapthread = (pthread_t *) malloc(sizeof(pthread_t) * numthreads);
    pthread_mutex_init(&power_lock, NULL);
    pthread_mutex_init(&print_lock, NULL);
    pthread_mutex_init(&power_pos_lock, NULL);

    // Imagemagick stuff
    Image *img = NULL;
    ImageInfo *img_info;
    ExceptionInfo exception;
    DrawInfo *di;

    InitializeMagick(*argv);
    GetExceptionInfo(&exception);
    img_info = CloneImageInfo((ImageInfo *) NULL);

    if (optind == argc) {
        fprintf(stderr, "FATAL:  Must provide at least one dump file.\n");
        Usage(argv[0]);
    }

    for (int x = optind; x < argc; x++) {
        if (ProcessGPSFile(argv[x]) < 0) {
            fprintf(stderr, "WARNING:  Unrecoverable error processing GPS data file \"%s\", skipping.\n",
                    argv[x]);
            //exit(1);
        }
    }

    fprintf(stderr, "Graphing %d sample points.\n",
            sample_points);

    map_avg_lat = (double) (global_map_avg.min_lat + global_map_avg.max_lat) / 2;
    map_avg_lon = (double) (global_map_avg.min_lon + global_map_avg.max_lon) / 2;

    // Fit the whole map
    scale = BestMapScale(global_map_avg.min_lat, global_map_avg.min_lon,
                         global_map_avg.max_lat, global_map_avg.max_lon);

    fprintf(stderr, "Map image scale: %ld\n", scale);
    fprintf(stderr, "Minimum Corner (lat/lon): %f x %f\n", global_map_avg.min_lat,
            global_map_avg.min_lon);
    fprintf(stderr, "Maximum Corner (lat/lon): %f x %f\n", global_map_avg.max_lat,
            global_map_avg.max_lon);
    fprintf(stderr, "Map center (lat/lon): %f x %f\n", map_avg_lat, map_avg_lon);

    if (user_scale != 0) {
        fprintf(stderr, "Overriding with custom user scale: %ld\n", user_scale);
        fprintf(stderr, "User map center (lat/lon): %f x %f\n", user_lat, user_lon);

        map_avg_lat = user_lat;
        map_avg_lon = user_lon;
        scale = user_scale;
    }

    if (scale == 0) {
        fprintf(stderr, "Unable to find a map at any scale to fit the data.\n");
        exit(0);
    }

    if (!usermap) {
        snprintf(mapname, 1024, "map_%f_%f_%ld_%d_%d.gif", map_avg_lat, map_avg_lon,
                 scale, map_width, map_height);
    }

    if (mapoutname[0] == '\0')
        snprintf(mapoutname, 1024, "map_%f_%f_%ld_%d_%d.png", map_avg_lat, map_avg_lon,
                 scale, map_width, map_height);

    printf("Loading map into Imagemagick structures.\n");
    strcpy(img_info->filename, mapname);
    img = ReadImage(img_info, &exception);

    if (img == (Image *) NULL) {
        if (usermap) {
            printf("Unable to load '%s'\n", mapname);
            exit(1);
        }

        char url[1024];
        snprintf(url, 1024, url_template, map_avg_lat, map_avg_lon, scale,
                 map_width, map_height,
                 metric ? "&DU=KM" : "");

        printf("Map url: %s\n", url);
        printf("Fetching map...\n");

        char geturl[1024];
        snprintf(geturl, 1024, download_template, url, mapname);
        system(geturl);

        printf("Loading map into Imagemagick structures.\n");
        strcpy(img_info->filename, mapname);
        img = ReadImage(img_info, &exception);
    }

    if (img == (Image *) NULL) {
        fprintf(stderr, "FATAL:  ImageMagick error:\n");
        MagickError(exception.severity,exception.reason,exception.description);
        exit(0);
    }

    strcpy(img_info->filename, mapoutname);
    strcpy(img->filename, mapoutname);

    di = CloneDrawInfo(img_info, NULL);

    // Convert it to greyscale and then back to color
    if (convert_greyscale) {
        fprintf(stderr, "Converting map to greyscale.\n");
        SetImageType(img, GrayscaleType);
        SetImageType(img, TrueColorType);
    }

    fprintf(stderr, "Calculating network coordinates and statistics...\n");
    ProcessNetData(verbose);

    fprintf(stderr, "Assigning network colors...\n");
    AssignNetColors();

    if (draw_track) {
        fprintf(stderr, "Drawing track coordinates...\n");
        DrawNetTracks(img, di);
    }

    if (draw_center) {
        fprintf(stderr, "Drawing center dot size %d...\n", center_dot);
        DrawNetCenterDot(img, di, center_dot);
    }

    if (draw_interpolated && power_data == 0) {
        fprintf(stderr, "ERROR:  Interpolated power drawing requested, but none of the GPS datafiles being\n"
                "processed have power data.  Not doing interpolated graphing.\n");
    } else if (draw_interpolated) {
        fprintf(stderr, "Drawing network power interpolations...\n");

        DrawNetPower(img, di, interpolation_res);
    }

    if (draw_range) {
        fprintf(stderr, "Calculating and drawing network circles...\n");
        DrawNetCircles(img, di);
    }

    if (draw_bounds) {
        fprintf(stderr, "Calculating and drawing bounding rectangles...\n");
        DrawNetBoundRects(img, di, 0);
    }

    WriteImage(img_info, img);

    DestroyDrawInfo(di);
    DestroyImage(img);

    DestroyMagick();

    delete mapthread;

    if (!keep_gif && !usermap) {
        fprintf(stderr, "Unlinking downloaded map.\n");
        unlink(mapname);
    }

}

#endif

